/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "execution_manager.h"

#include <cstddef>
#include <fstream>
#include <mutex>
#include <sstream>

#include "common/common.h"
#include "common/config.h"
#include "common/log.h"
#include "errors.h"
#include "executor_delete.h"
#include "executor_index_scan.h"
#include "executor_insert.h"
#include "executor_nestedloop_join.h"
#include "executor_projection.h"
#include "executor_seq_scan.h"
#include "executor_update.h"
#include "index/ix.h"
#include "index/ix_index_handle.h"
#include "optimizer/plan.h"
#include "parser/ast.h"
#include "record/rm_defs.h"
#include "record_printer.h"
#include "transaction/transaction.h"
#include "transaction/transaction_manager.h"

bool QlManager::need_output = true;

const char *help_info =
    "Supported SQL syntax:\n"
    "  command ;\n"
    "command:\n"
    "  CREATE TABLE table_name (column_name type [, column_name type ...])\n"
    "  DROP TABLE table_name\n"
    "  CREATE INDEX table_name (column_name)\n"
    "  DROP INDEX table_name (column_name)\n"
    "  INSERT INTO table_name VALUES (value [, value ...])\n"
    "  DELETE FROM table_name [WHERE where_clause]\n"
    "  UPDATE table_name SET column_name = value [, column_name = value ...] [WHERE where_clause]\n"
    "  SELECT selector FROM table_name [WHERE where_clause]\n"
    "type:\n"
    "  {INT | FLOAT | CHAR(n)}\n"
    "where_clause:\n"
    "  condition [AND condition ...]\n"
    "condition:\n"
    "  column op {column | value}\n"
    "column:\n"
    "  [table_name.]column_name\n"
    "op:\n"
    "  {= | <> | < | > | <= | >=}\n"
    "selector:\n"
    "  {* | column [, column ...]}\n";

// 主要负责执行DDL语句
void QlManager::run_mutli_query(std::shared_ptr<Plan> plan, Context *context) {
  if (auto x = std::dynamic_pointer_cast<DDLPlan>(plan)) {
    switch (x->tag) {
      case T_CreateTable: {
        sm_manager_->create_table(x->tab_name_, x->cols_, context);
        break;
      }
      case T_DropTable: {
        sm_manager_->drop_table(x->tab_name_, context);
        break;
      }
      case T_CreateIndex: {
        sm_manager_->create_index(x->tab_name_, x->tab_col_names_, context);
        break;
      }
      case T_DropIndex: {
        sm_manager_->drop_index(x->tab_name_, x->tab_col_names_, context);
        break;
      }
      default:
        throw InternalError("Unexpected field type");
        break;
    }
  }
}

// 执行help; show tables; desc table; begin; commit; abort;语句
void QlManager::run_cmd_utility(std::shared_ptr<Plan> plan, txn_id_t *txn_id, Context *context) {
  if (auto x = std::dynamic_pointer_cast<OtherPlan>(plan)) {
    switch (x->tag) {
      case T_Help: {
        memcpy(context->data_send_ + *(context->offset_), help_info, strlen(help_info));
        *(context->offset_) = strlen(help_info);
      } break;

      case T_LoadTable: {
        if (!sm_manager_->get_dm()->is_file(x->tab_name_)) {
          throw InternalError("table not exists.");
        }
        std::fstream file(x->path_name_);

        if (!file.is_open()) {
          throw InternalError("file not exists.");
        }

        auto &tab_meta = sm_manager_->db_.get_table(x->tab_name_);

        size_t n = 0;
        std::string str_line;

        {
          std::string col_name;
          // 检查表头。
          std::getline(file, str_line);
          std::stringstream str_stream(str_line);

          for (size_t i = 0, m = tab_meta.cols.size(); i < m; ++i) {
            std::getline(str_stream, col_name, ',');
            ++n;
            // std::cout << col_name << " : " << tab_meta.cols[i].name << std::endl;
            // ASSERT(col_name == tab_meta.cols[i].name, "col_name not equal tab_meta.col.name");
          }

          // ASSERT(n == tab_meta.cols.size(), "col number not equal.");
        }

        auto tab = sm_manager_->fhs_.at(x->tab_name_).get();
        std::vector<Value> values(n);

        for (size_t i = 0; i < n; ++i) {
          values[i].type = tab_meta.cols[i].type;
        }

        int record_size = tab->get_file_hdr().record_size;
        RmRecord rcd(record_size);

        std::string str_col;

        while (std::getline(file, str_line)) {
          std::stringstream str_stream(str_line);

          for (size_t i = 0; i < n; ++i) {
            std::getline(str_stream, str_col, ',');
            str_col >> values[i];

            values[i].resize(tab_meta.cols[i].len);
            memcpy(rcd.data + tab_meta.cols[i].offset, values[i].data(), tab_meta.cols[i].len);
          }

          Rid rid = tab->insert_record(rcd.data, nullptr);

          for (auto &ix_meta : tab_meta.indexes) {
            auto ix =
                sm_manager_->ihs_
                    .at(sm_manager_->get_ix_manager()->get_index_name(tab_meta.name, ix_meta.cols))
                    .get();
            RmRecord index_rcd(ix_meta.col_tot_len);
            ix_meta.make_index_tuple(&rcd, &index_rcd);
            ix->insert_entry(index_rcd.data, rid, nullptr);
          }
        }

        sm_manager_->get_bpm()->flush_all_pages(tab->GetFd());
      } break;

      case T_CreateStaticCheckPoint: {
        TransactionManager::allow_begin = false;

        std::unique_lock<std::mutex> lock(TransactionManager::check_point_latch);

        while (TransactionManager::ATT.size() > 1) {
          TransactionManager::wait_cv.wait(lock);
        }

        sm_manager_->get_bpm()->flush_all_pages();
        context->log_mgr_->close_loop();
        context->log_mgr_->flush_log_to_disk();
        auto disk_manager = sm_manager_->get_dm();
        disk_manager->destroy_log();
        disk_manager->create_file(LOG_FILE_NAME);

        context->log_mgr_->begin_flush_loop();

        TransactionManager::allow_begin = true;
        TransactionManager::begin_cv.notify_all();
      } break;

      case T_ShowTable: {
        sm_manager_->show_tables(context);
        break;
      }

      case T_ShowIndex: {
        sm_manager_->show_index(x->tab_name_, context);
        break;
      }
      case T_DescTable: {
        sm_manager_->desc_table(x->tab_name_, context);
        break;
      }
      case T_Transaction_begin: {
        // 显示开启一个事务
        // auto txn = context->txn_;
        // BeginLogRecord log_record(txn->get_transaction_id());

        // 这个时候再写入begin日志。
        // log_record.prev_lsn_ = txn->get_prev_lsn();
        // lsn_t prev_lsn = context->log_mgr_->add_log_to_buffer(&log_record);
        // txn->set_prev_lsn(prev_lsn);
        context->txn_->set_txn_mode(EXPLICIT_TXN);
        break;
      }
      case T_Transaction_commit: {
        context->txn_ = txn_mgr_->get_transaction(*txn_id);

        // 无论是否是隐性事务，都需要释放锁等操作，只是需要修改日志
        txn_mgr_->commit(context->txn_, context->log_mgr_);
        break;
      }

      // 所以rollabck不是等于abort，那为什么在语法分析的阶段还要分开？有毛？
      case T_Transaction_rollback: {
        context->txn_ = txn_mgr_->get_transaction(*txn_id);
        txn_mgr_->abort(context->txn_, context->log_mgr_);
        break;
      }

      case T_Transaction_abort: {
        context->txn_ = txn_mgr_->get_transaction(*txn_id);
        txn_mgr_->abort(context->txn_, context->log_mgr_);
        break;
      }
      default:
        throw InternalError("Unexpected field type");
        break;
    }

  } else if (auto x = std::dynamic_pointer_cast<SetKnobPlan>(plan)) {
    switch (x->set_knob_type_) {
      case ast::SetKnobType::EnableNestLoop: {
        planner_->set_enable_nestedloop_join(x->bool_value_);
        break;
      }
      case ast::SetKnobType::EnableSortMerge: {
        planner_->set_enable_sortmerge_join(x->bool_value_);
        break;
      }
      case ast::SetKnobType::EnableOutFile: {
        need_output = x->bool_value_;
        break;
      }
      default: {
        throw RMDBError("Not implemented!\n");
        break;
      }
    }
  }
}

// 执行select语句，select语句的输出除了需要返回客户端外，还需要写入output.txt文件中
void QlManager::select_from(std::unique_ptr<AbstractExecutor> executorTreeRoot,
                            std::vector<TabCol> sel_cols, Context *context) {
  executorTreeRoot->beginTuple();  // 会进行类型等等检查，所以先执行这个。

  std::vector<std::string> captions;
  captions.reserve(sel_cols.size());

  bool need_tab_name = false;

  //// 如果是多表，就把表名也打印出来。 不过现在看来是画蛇添足了。
  // {
  //   std::string tab_name;

  //   for (auto &sel_col : sel_cols) {
  //     if (tab_name.empty()) {
  //       tab_name = sel_col.tab_name;
  //     } else if (tab_name != sel_col.tab_name) {
  //       need_tab_name = true;
  //       break;
  //     }
  //   }
  // }

  for (auto &sel_col : sel_cols) {
    if (!sel_col.alias.empty()) {
      captions.push_back(sel_col.alias);
    } else {
      captions.push_back(get_column_name(need_tab_name ? sel_col.tab_name : "", sel_col.col_name,
                                         sel_col.aggregation_type));
    }
  }

  // Print header into buffer
  RecordPrinter rec_printer(sel_cols.size());
  rec_printer.print_separator(context);
  rec_printer.print_record(captions, context);
  rec_printer.print_separator(context);
  // print header into file
  std::fstream outfile;

  if (need_output) {
    outfile.open("output.txt", std::ios::out | std::ios::app);
    outfile << "|";
    for (size_t i = 0; i < captions.size(); ++i) {
      outfile << " " << captions[i] << " |";
    }
    outfile << "\n";
  }

  // Print records
  size_t num_rec = 0;
  // 执行query_plan
  for (; !executorTreeRoot->is_end(); executorTreeRoot->nextTuple()) {
    auto Tuple = executorTreeRoot->Next();
    std::vector<std::string> columns;
    for (auto &col : executorTreeRoot->cols()) {
      std::string col_str;
      char *rec_buf = Tuple->data + col.offset;
      if (col.type == TYPE_INT) {
        col_str = std::to_string(*(int *)rec_buf);
      } else if (col.type == TYPE_FLOAT) {
        col_str = std::to_string(*(float *)rec_buf);
      } else if (col.type == TYPE_STRING) {
        col_str = std::string((char *)rec_buf, col.len);
        col_str.resize(strlen(col_str.c_str()));
      }
      columns.push_back(col_str);
    }
    // print record into buffer
    rec_printer.print_record(columns, context);

    if (need_output) {
      // print record into file
      outfile << "|";
      for (size_t i = 0; i < columns.size(); ++i) {
        outfile << " " << columns[i] << " |";
      }
      outfile << "\n";
    }

    num_rec++;
  }

  if (need_output) {
    outfile.close();
  }
  // Print footer into buffer
  rec_printer.print_separator(context);
  // Print record count into buffer
  RecordPrinter::print_record_count(num_rec, context);
}

// 执行DML语句
void QlManager::run_dml(std::unique_ptr<AbstractExecutor> exec) { exec->Next(); }