/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cstddef>

#include "common/common.h"
#include "common/context.h"
#include "common/log.h"
#include "defs.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "record/rm_defs.h"
#include "record/rm_scan.h"
#include "system/sm.h"
#include "system/sm_meta.h"

class SeqScanExecutor : public AbstractExecutor {
 private:
  std::string tab_name_;              // 表的名称
  std::vector<Condition> conds_;      // scan的条件
  RmFileHandle *fh_;                  // 表的数据文件句柄
  std::vector<ColMeta> cols_;         // scan后生成的记录的字段
  size_t len_;                        // scan后生成的每条记录的长度
  std::vector<Condition> fed_conds_;  // 同conds_，两个字段相同

  Rid rid_;
  std::unique_ptr<RecScan> scan_;  // table_iterator

  SmManager *sm_manager_;

 public:
  // 扫描算子需要读操作，所以需要context来获取事务信息。
  SeqScanExecutor(SmManager *sm_manager, std::string tab_name, std::vector<Condition> conds,
                  Context *context) {
    sm_manager_ = sm_manager;
    tab_name_ = std::move(tab_name);
    TabMeta &tab = sm_manager_->db_.get_table(tab_name_);
    fh_ = sm_manager_->fhs_.at(tab_name_).get();
    cols_ = tab.cols;
    len_ = cols_.back().offset + cols_.back().len;

    AbstractExecutor::context_ = context;

    for (auto &cond : conds) {
      // 现在条件的左表必然是tab_name_.
      // 现在左右表名不等，条件为等值，可以认为是归并连接。
      // 归并连接条件不可用作扫描算子的筛选。
      if (cond.op == OP_EQ && cond.is_rhs_column &&
          cond.lhs_col.tab_name != cond.rhs_col.tab_name) {
        ;  // 什么都不做。
      } else {
        fed_conds_.push_back(cond);  //
      }
    }

    conds_ = fed_conds_;
  }

  void beginTuple() override {
    ASSERT(context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd()),
           "get lock failed in seq scan.");

    open_conditions(fed_conds_);

    scan_ = std::make_unique<RmScan>(fh_);
    while (!scan_->is_end()) {
      auto rm = fh_->get_record(scan_->rid(), context_);

      if (!check_conditions(rm.get(), rm.get(), fed_conds_, cols_, cols_)) {
        scan_->next();
      } else {
        rid_ = scan_->rid();
        // 读取的记录放行级读锁
        ASSERT(context_->lock_mgr_->lock_shared_on_record(context_->txn_, rid_, fh_->GetFd()),
               "get rec lock failed in select.");
        break;
      }
    }
  }

  void nextTuple() override {
    for (scan_->next(); !scan_->is_end(); scan_->next()) {
      auto rcd = fh_->get_record(scan_->rid(), context_);

      if (check_conditions(rcd.get(), rcd.get(), fed_conds_, cols_, cols_)) {
        rid_ = scan_->rid();
        // 读取的记录放行级读锁
        ASSERT(context_->lock_mgr_->lock_shared_on_record(context_->txn_, rid_, fh_->GetFd()),
               "get rec lock failed in select.");
        break;
      }
    }
  }

  std::unique_ptr<RmRecord> Next() override { return fh_->get_record(rid_, context_); }

  bool is_end() const override { return scan_->is_end(); }
  Rid &rid() override { return rid_; }
  size_t tupleLen() const override { return len_; }

  const std::vector<ColMeta> &cols() const override { return cols_; }
};