/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <unistd.h>

#include <cstddef>
#include <memory>

#include "common/common.h"
#include "common/log.h"
#include "execution/executor_index_scan.h"
#include "execution/executor_seq_scan.h"
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "record/rm_defs.h"
#include "record_printer.h"
#include "system/sm.h"

enum class Table_scn_type { LeftTbl, RightTbl, SeqScan, IndexScan };

class SortMergeJoinExecutor : public AbstractExecutor {
 private:
  std::unique_ptr<AbstractExecutor> left_;            // 左儿子节点（需要join的表）
  std::unique_ptr<AbstractExecutor> right_;           // 右儿子节点（需要join的表）
  size_t len_;                                        // join后获得的每条记录的长度
  std::vector<ColMeta> cols_, cols_left, cols_right;  // join后获得的记录的字段
  ColMeta col_set_left, col_set_right;

  std::vector<Condition> fed_conds_;        // join条件
  std::map<TabCol, Value> prev_feed_dict_;  // 字段到值的映射
  std::unique_ptr<RmRecord> current_rec_;
  std::unique_ptr<RmRecord> left_rec_;  // 把左边的tuple存下来可以少调用很多次left_->Next();
  std::vector<std::unique_ptr<RmRecord>> tuples_left, tuples_right;  // 左右边全部数据，用于排序
  std::vector<std::unique_ptr<RmRecord>>::iterator current_tuple_left, current_tuple_right;
  bool need_reverse_;  // 是否需要调整中间排序的结果顺序
  bool need_sleep_;

 public:
  SortMergeJoinExecutor(std::unique_ptr<AbstractExecutor> left,
                        std::unique_ptr<AbstractExecutor> right, std::vector<Condition> conds,
                        bool need_reverse, bool need_sleep = false) {
    need_sleep_ = need_sleep;
    left_ = std::move(left);
    right_ = std::move(right);
    need_reverse_ = need_reverse;
    len_ = left_->tupleLen() + right_->tupleLen();
    cols_ = left_->cols();
    cols_left = left_->cols();
    cols_right = right_->cols();
    auto right_cols = right_->cols();

    for (auto &col : right_cols) {
      col.offset += left_->tupleLen();
    }

    cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
    fed_conds_ = std::move(conds);

    col_set_left = *get_col(cols_left, (fed_conds_[0].lhs_col.tab_name == cols_left[0].tab_name)
                                           ? fed_conds_[0].lhs_col
                                           : fed_conds_[0].rhs_col);

    col_set_right = *get_col(cols_right, (fed_conds_[0].rhs_col.tab_name == cols_right[0].tab_name)
                                             ? fed_conds_[0].rhs_col
                                             : fed_conds_[0].lhs_col);
  }
  Table_scn_type table_san_type(Table_scn_type table_scan) {
    SeqScanExecutor *derived1Ptr;
    IndexScanExecutor *derived2Ptr;
    switch (table_scan) {
      case Table_scn_type::LeftTbl:
        derived1Ptr = dynamic_cast<SeqScanExecutor *>(left_.get());
        derived2Ptr = dynamic_cast<IndexScanExecutor *>(left_.get());
        break;
      case Table_scn_type::RightTbl:
        derived1Ptr = dynamic_cast<SeqScanExecutor *>(right_.get());
        derived2Ptr = dynamic_cast<IndexScanExecutor *>(right_.get());
        break;
      default:
        break;
    }
    // 根据转换结果判断类型
    if (derived1Ptr) {
      return Table_scn_type::SeqScan;
    } else if (derived2Ptr) {
      return Table_scn_type::IndexScan;
    }
  }

  void beginTuple() override {
    // if (need_sleep_) {
    //   sleep(1);
    // }

    open_conditions(fed_conds_);
    // 左表
    std::vector<std::string> captions, captions_right;
    captions.reserve(cols_left.size());
    for (auto &sel_col : cols_left) {
      captions.push_back(sel_col.name);
    }
    RecordPrinter rec_printer(cols_left.size()), rec_printer_right(cols_right.size());
    // print header into file
    std::fstream outfile;
    outfile.open("sorted_results.txt", std::ios::out | std::ios::app);
    for (left_->beginTuple(); !left_->is_end(); left_->nextTuple()) {
      auto tuple = left_->Next();
      tuples_left.emplace_back(std::move(tuple));
    }

    if (table_san_type(Table_scn_type::LeftTbl) == Table_scn_type::SeqScan) {
      Value lhs;
      Value rhs;
      lhs.type = col_set_left.type;
      rhs.type = col_set_left.type;
      auto cmp = [this, &lhs, &rhs](std::unique_ptr<RmRecord> &left,
                                    std::unique_ptr<RmRecord> &right) {
        lhs.set_data(left->data + col_set_left.offset, col_set_left.len);
        rhs.set_data(right->data + col_set_left.offset, col_set_left.len);

        CompResult res = lhs.compare_value(rhs);
        return res == CompResult::LT;
      };
      std::sort(tuples_left.begin(), tuples_left.end(), cmp);
    }
    // 右表
    captions_right.reserve(cols_right.size());
    for (auto &sel_col : cols_right) {
      captions_right.push_back(sel_col.name);
    }
    for (right_->beginTuple(); !right_->is_end(); right_->nextTuple()) {
      auto tuple = right_->Next();
      tuples_right.emplace_back(std::move(tuple));
    }
    if (table_san_type(Table_scn_type::RightTbl) == Table_scn_type::SeqScan) {
      Value lhs_right;
      Value rhs_right;
      lhs_right.type = col_set_right.type;
      rhs_right.type = col_set_right.type;
      auto cmp_right = [this, &lhs_right, &rhs_right](std::unique_ptr<RmRecord> &left,
                                                      std::unique_ptr<RmRecord> &right) {
        lhs_right.set_data(left->data + col_set_right.offset, col_set_right.len);
        rhs_right.set_data(right->data + col_set_right.offset, col_set_right.len);

        CompResult res = lhs_right.compare_value(rhs_right);
        return res == CompResult::LT;
      };
      std::sort(tuples_right.begin(), tuples_right.end(), cmp_right);
    }

    if (need_reverse_ == true) {
      // 左表表头
      outfile << "|";
      for (size_t i = 0; i < captions.size(); ++i) {
        outfile << " " << captions[i] << " |";
      }
      outfile << "\n";
      for (auto &tuple : tuples_left) {
        std::vector<std::string> columns;
        for (auto &col : cols_left) {
          std::string col_str_2;
          char *rec_buf = tuple->data + col.offset;
          if (col.type == TYPE_INT) {
            col_str_2 = std::to_string(*(int *)rec_buf);
          } else if (col.type == TYPE_FLOAT) {
            col_str_2 = std::to_string(*(float *)rec_buf);
          } else if (col.type == TYPE_STRING) {
            col_str_2 = std::string((char *)rec_buf, col.len);
            col_str_2.resize(strlen(col_str_2.c_str()));
          }
          columns.push_back(col_str_2);
        }
        outfile << "|";
        for (size_t i = 0; i < columns.size(); ++i) {
          outfile << " " << columns[i] << " |";
        }
        outfile << "\n";
      }
      // 右表表头
      outfile << "|";
      for (size_t i = 0; i < captions_right.size(); ++i) {
        outfile << " " << captions_right[i] << " |";
      }
      outfile << "\n";
      for (auto &tuple : tuples_right) {
        std::vector<std::string> columns;
        for (auto &col : cols_right) {
          std::string col_str;
          char *rec_buf = tuple->data + col.offset;
          if (col.type == TYPE_INT) {
            col_str = std::to_string(*(int *)rec_buf);
          } else if (col.type == TYPE_FLOAT) {
            col_str = std::to_string(*(float *)rec_buf);
          } else if (col.type == TYPE_STRING) {
            col_str = std::string((char *)rec_buf, col.len);
            col_str.resize(strlen(col_str.c_str()));
          }
          columns.push_back(col_str);
        }
        outfile << "|";
        for (size_t i = 0; i < columns.size(); ++i) {
          outfile << " " << columns[i] << " |";
        }
        outfile << "\n";
      }
    } else {
      // 右表表头
      outfile << "|";
      for (size_t i = 0; i < captions_right.size(); ++i) {
        outfile << " " << captions_right[i] << " |";
      }
      outfile << "\n";
      for (auto &tuple : tuples_right) {
        std::vector<std::string> columns;
        for (auto &col : cols_right) {
          std::string col_str;
          char *rec_buf = tuple->data + col.offset;
          if (col.type == TYPE_INT) {
            col_str = std::to_string(*(int *)rec_buf);
          } else if (col.type == TYPE_FLOAT) {
            col_str = std::to_string(*(float *)rec_buf);
          } else if (col.type == TYPE_STRING) {
            col_str = std::string((char *)rec_buf, col.len);
            col_str.resize(strlen(col_str.c_str()));
          }
          columns.push_back(col_str);
        }
        outfile << "|";
        for (size_t i = 0; i < columns.size(); ++i) {
          outfile << " " << columns[i] << " |";
        }
        outfile << "\n";
      }
      // 左表表头
      outfile << "|";
      for (size_t i = 0; i < captions.size(); ++i) {
        outfile << " " << captions[i] << " |";
      }
      outfile << "\n";
      for (auto &tuple : tuples_left) {
        std::vector<std::string> columns;
        for (auto &col : cols_left) {
          std::string col_str_2;
          char *rec_buf = tuple->data + col.offset;
          if (col.type == TYPE_INT) {
            col_str_2 = std::to_string(*(int *)rec_buf);
          } else if (col.type == TYPE_FLOAT) {
            col_str_2 = std::to_string(*(float *)rec_buf);
          } else if (col.type == TYPE_STRING) {
            col_str_2 = std::string((char *)rec_buf, col.len);
            col_str_2.resize(strlen(col_str_2.c_str()));
          }
          columns.push_back(col_str_2);
        }
        outfile << "|";
        for (size_t i = 0; i < columns.size(); ++i) {
          outfile << " " << columns[i] << " |";
        }
        outfile << "\n";
      }
    }
    outfile.close();
    current_tuple_left = tuples_left.begin();
    current_tuple_right = tuples_right.begin();

    make_join();
  }

  void nextTuple() override {
    Value lhs, rhs;
    lhs.type = col_set_left.type;
    rhs.type = col_set_right.type;
    RmRecord *left = current_tuple_left->get();    // 直接获取RmRecord指针
    RmRecord *right = current_tuple_right->get();  // 直接获取RmRecord指针
    lhs.set_data(left->data + col_set_left.offset, col_set_left.len);
    rhs.set_data(right->data + col_set_right.offset, col_set_right.len);
    CompResult res = lhs.compare_value(rhs);
    if (res == CompResult::LT) {
      if (current_tuple_left != tuples_left.end()) ++current_tuple_left;
    } else {
      if (current_tuple_right != tuples_right.end()) ++current_tuple_right;
    }
    if (current_tuple_left == tuples_left.end() || current_tuple_right == tuples_right.end()) {
      current_tuple_left = tuples_left.end();
      current_tuple_right = tuples_right.end();
      return;
    }
    make_join();
  }

  // 这里Next不应该move，不然如果多次调用Next的话，后面会出现错误.
  // std::unique_ptr<RmRecord> Next() override { return std::move(current_rec_); }

  // 把拷贝的一份返回出去。
  std::unique_ptr<RmRecord> Next() override { return std::make_unique<RmRecord>(*current_rec_); }
  bool is_end_test() { return current_tuple_left == tuples_left.end(); }
  bool is_end() const override {
    // return current_tuple_left == tuples_left.end();
    bool is_left_end = current_tuple_left == tuples_left.end();
    bool is_right_end = current_tuple_right == tuples_right.end();
    return is_left_end || is_right_end;
  }

  Rid &rid() override { return _abstract_rid; }

  const std::vector<ColMeta> &cols() const override { return cols_; };

  void make_join_test() {
    while (true) {  // 因为每次调用Next都会消耗资源（会重新分配RmRecord）所以尽量少调用Next
      //  std::unique_ptr<RmRecord> right_rec = std::make_unique<RmRecord>(**current_tuple_right);
      RmRecord *left = current_tuple_left->get();    // 直接获取RmRecord指针
      RmRecord *right = current_tuple_right->get();  // 直接获取RmRecord指针
      if (check_conditions(left, right, fed_conds_, left_->cols(), right_->cols())) {
        current_rec_ = std::make_unique<RmRecord>(len_);
        size_t left_len = left_->tupleLen();
        memcpy(current_rec_->data, left->data, left_len);
        memcpy(current_rec_->data + left_len, right->data, right_->tupleLen());
        break;
      } else {
        ++current_tuple_right;
        if (current_tuple_right == tuples_right.end()) {
          ++current_tuple_left;
          //  left_rec_ = std::make_unique<RmRecord>(**current_tuple_left);
          current_tuple_right = tuples_right.begin();
        }
        if (is_end_test()) {
          break;
        }
      }
    }
  }

  void make_join() {
    while (true) {  // 因为每次调用Next都会消耗资源（会重新分配RmRecord）所以尽量少调用Next
      RmRecord *left = current_tuple_left->get();    // 直接获取RmRecord指针
      RmRecord *right = current_tuple_right->get();  // 直接获取RmRecord指针
      if (check_conditions(left, right, fed_conds_, left_->cols(), right_->cols())) {
        current_rec_ = std::make_unique<RmRecord>(len_);
        size_t left_len = left_->tupleLen();
        memcpy(current_rec_->data, left->data, left_len);
        memcpy(current_rec_->data + left_len, right->data, right_->tupleLen());
        break;
      } else {
        // 移动指针，准备检查下一个记录对
        Value lhs, rhs;
        lhs.type = col_set_left.type;
        rhs.type = col_set_right.type;
        lhs.set_data(left->data + col_set_left.offset, col_set_left.len);
        rhs.set_data(right->data + col_set_right.offset, col_set_right.len);
        CompResult res = lhs.compare_value(rhs);
        if (res == CompResult::LT) {
          if (current_tuple_left != tuples_left.end()) ++current_tuple_left;
        } else {
          if (current_tuple_right != tuples_right.end()) ++current_tuple_right;
        }
        if (current_tuple_left == tuples_left.end() || current_tuple_right == tuples_right.end())
          break;
      }
    }
  }

  size_t tupleLen() const override { return len_; }

  void make_tuple(std::unique_ptr<RmRecord> &left_tuple, std::unique_ptr<RmRecord> &right_tuple) {
    current_rec_ = std::make_unique<RmRecord>(len_);
    size_t left_len = left_->tupleLen();
    memcpy(current_rec_->data, left_tuple->data, left_len);
    memcpy(current_rec_->data + left_len, right_tuple->data, right_->tupleLen());

#ifdef DEBUG
    printf("left_tuple : ");
    for (auto &left_col : left_->cols()) {
      Value val;
      val.type = left_col.type;
      val.set_data(left_tuple->data + left_col.offset, left_col.len);

      printf("%s ", val.to_string().c_str());
    }

    printf(" right_tuple : ");

    for (auto &right_col : right_->cols()) {
      Value val;
      val.type = right_col.type;
      val.set_data(right_tuple->data + right_col.offset, right_col.len);

      printf("%s ", val.to_string().c_str());
    }

    printf(" tuple : ");
    for (auto &col : cols_) {
      Value val;
      val.type = col.type;
      val.set_data(current_rec_->data + col.offset, col.len);

      printf("%s ", val.to_string().c_str());
    }
    printf("\n");
#endif
  }
};