/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cerrno>
#include <cstring>
#include <memory>
#include <string>

#include "common/common.h"
#include "common/context.h"
#include "errors.h"
#include "execution/execution_aggregation_and_group.h"
#include "execution/execution_sort.h"
#include "execution/executor_abstract.h"
#include "execution/executor_delete.h"
#include "execution/executor_index_scan.h"
#include "execution/executor_insert.h"
#include "execution/executor_nestedloop_join.h"
#include "execution/executor_projection.h"
#include "execution/executor_seq_scan.h"
#include "execution/executor_sortmerge_join.h"
#include "execution/executor_update.h"
#include "optimizer/plan.h"

typedef enum portalTag {
  PORTAL_Invalid_Query = 0,
  PORTAL_ONE_SELECT,
  PORTAL_DML_WITHOUT_SELECT,
  PORTAL_MULTI_QUERY,
  PORTAL_CMD_UTILITY
} portalTag;

struct PortalStmt {
  portalTag tag;

  std::vector<TabCol> sel_cols;
  std::unique_ptr<AbstractExecutor> root;
  std::shared_ptr<Plan> plan;

  PortalStmt(portalTag tag_, std::vector<TabCol> sel_cols_, std::unique_ptr<AbstractExecutor> root_,
             std::shared_ptr<Plan> plan_)
      : tag(tag_), sel_cols(std::move(sel_cols_)), root(std::move(root_)), plan(std::move(plan_)) {}
};

class Portal {
 private:
  SmManager *sm_manager_;

 public:
  Portal(SmManager *sm_manager) : sm_manager_(sm_manager) {}
  ~Portal() {}

  // 将查询执行计划转换成对应的算子树
  std::shared_ptr<PortalStmt> start(std::shared_ptr<Plan> plan, Context *context) {
    // 这里可以将select进行拆分，例如：一个select，带有return的select等
    if (auto x = std::dynamic_pointer_cast<OtherPlan>(plan)) {
      return std::make_shared<PortalStmt>(PORTAL_CMD_UTILITY, std::vector<TabCol>(),
                                          std::unique_ptr<AbstractExecutor>(), plan);
    } else if (auto x = std::dynamic_pointer_cast<SetKnobPlan>(plan)) {
      return std::make_shared<PortalStmt>(PORTAL_CMD_UTILITY, std::vector<TabCol>(),
                                          std::unique_ptr<AbstractExecutor>(), plan);
    } else if (auto x = std::dynamic_pointer_cast<DDLPlan>(plan)) {
      return std::make_shared<PortalStmt>(PORTAL_MULTI_QUERY, std::vector<TabCol>(),
                                          std::unique_ptr<AbstractExecutor>(), plan);
    } else if (auto x = std::dynamic_pointer_cast<DMLPlan>(plan)) {
      switch (x->tag) {
        case T_select: {
          std::shared_ptr<ProjectionPlan> p =
              std::dynamic_pointer_cast<ProjectionPlan>(x->subplan_);

          check_sub_query(x->subplan_, context);

          std::unique_ptr<AbstractExecutor> root = convert_plan_executor(p, context);
          return std::make_shared<PortalStmt>(PORTAL_ONE_SELECT, std::move(p->sel_cols_),
                                              std::move(root), plan);
        }

        case T_Update: {
          // 更新和删除，这个地方会先使用scan算子获得要删除或者更新的rid，
          // 这个时候如果直接上锁，不太符合想要的 “上锁全部放在执行阶段”
          // ，所以把算子放进了更新和删除的算子中。 在begintuple中使用。
          check_sub_query(x->subplan_, context);

          std::unique_ptr<AbstractExecutor> scan = convert_plan_executor(x->subplan_, context);

          std::unique_ptr<AbstractExecutor> root = std::make_unique<UpdateExecutor>(
              sm_manager_, x->tab_name_, x->set_clauses_, x->conds_, std::move(scan), context);
          return std::make_shared<PortalStmt>(PORTAL_DML_WITHOUT_SELECT, std::vector<TabCol>(),
                                              std::move(root), plan);
        }

        case T_Delete: {
          check_sub_query(x->subplan_, context);

          std::unique_ptr<AbstractExecutor> scan = convert_plan_executor(x->subplan_, context);

          std::unique_ptr<AbstractExecutor> root = std::make_unique<DeleteExecutor>(
              sm_manager_, x->tab_name_, x->conds_, std::move(scan), context);

          return std::make_shared<PortalStmt>(PORTAL_DML_WITHOUT_SELECT, std::vector<TabCol>(),
                                              std::move(root), plan);
        }

        case T_Insert: {
          std::unique_ptr<AbstractExecutor> root =
              std::make_unique<InsertExecutor>(sm_manager_, x->tab_name_, x->values_, context);

          return std::make_shared<PortalStmt>(PORTAL_DML_WITHOUT_SELECT, std::vector<TabCol>(),
                                              std::move(root), plan);
        }

        default:
          throw InternalError("Unexpected field type");
          break;
      }
    } else {
      throw InternalError("Unexpected field type");
    }
    return nullptr;
  }

  // 遍历算子树并执行算子生成执行结果
  void run(std::shared_ptr<PortalStmt> portal, QlManager *ql, txn_id_t *txn_id, Context *context) {
    switch (portal->tag) {
      case PORTAL_ONE_SELECT: {
        ql->select_from(std::move(portal->root), std::move(portal->sel_cols), context);
        break;
      }

      case PORTAL_DML_WITHOUT_SELECT: {
        ql->run_dml(std::move(portal->root));
        break;
      }

      case PORTAL_MULTI_QUERY: {
        ql->run_mutli_query(portal->plan, context);
        break;
      }
      case PORTAL_CMD_UTILITY: {
        ql->run_cmd_utility(portal->plan, txn_id, context);
        break;
      }
      default: {
        throw InternalError("Unexpected field type");
      }
    }
  }

  // 清空资源
  void drop() {}

  void check_conditions(std::vector<Condition> &conditions, Context *context) {
    for (auto &cond : conditions) {
      if (!cond.is_sub_query) {
        continue;
      }

      auto x = std::dynamic_pointer_cast<DMLPlan>(cond.sub_query_plan);

      if (x == nullptr || x->tag != T_select) {
        throw InternalError("sub query must be a select sentence");
      }

      check_sub_query(x->subplan_, context);
      std::shared_ptr<ProjectionPlan> p = std::dynamic_pointer_cast<ProjectionPlan>(x->subplan_);
      std::unique_ptr<AbstractExecutor> root = convert_plan_executor(p, context);
      cond.sub_query_executor = std::move(root);
    }
  }

  // 检查一整个语法树。存在递归。
  // 递归一： 遍历整个语法树
  // 递归二： 会遍历子语法树。
  void check_sub_query(std::shared_ptr<Plan> &plan, Context *context) {
    if (plan == nullptr) {
      return;
    }

    switch (plan->tag) {
      case T_SeqScan:
      case T_IndexScan: {
        auto x = std::dynamic_pointer_cast<ScanPlan>(plan);
        check_conditions(x->conds_, context);
        check_conditions(x->fed_conds_, context);
      } break;

      case T_NestLoop:
      case T_SortMerge: {
        auto x = std::dynamic_pointer_cast<JoinPlan>(plan);
        check_conditions(x->conds_, context);

        check_sub_query(x->left_, context);
        check_sub_query(x->right_, context);
      } break;

      case T_Projection: {
        auto x = std::dynamic_pointer_cast<ProjectionPlan>(plan);
        check_sub_query(x->subplan_, context);
      } break;

      case T_Sort: {
        auto x = std::dynamic_pointer_cast<SortPlan>(plan);
        check_sub_query(x->subplan_, context);
      } break;

      case T_AggregationAndGroup: {
        auto x = std::dynamic_pointer_cast<AggregationAndGroupPlan>(plan);
        check_conditions(x->having_conds_, context);
        check_sub_query(x->subplan_, context);
      } break;

      default: {
        return;
      } break;
    }
  }

  std::unique_ptr<AbstractExecutor> convert_plan_executor(std::shared_ptr<Plan> plan,
                                                          Context *context) {
    if (auto x = std::dynamic_pointer_cast<ProjectionPlan>(plan)) {
      return std::make_unique<ProjectionExecutor>(convert_plan_executor(x->subplan_, context),
                                                  x->sel_cols_);
    } else if (auto x = std::dynamic_pointer_cast<ScanPlan>(plan)) {
      if (x->tag == T_SeqScan) {
        return std::make_unique<SeqScanExecutor>(sm_manager_, x->tab_name_, x->conds_, context);
      } else {
        return std::make_unique<IndexScanExecutor>(sm_manager_, x->tab_name_, x->conds_,
                                                   x->index_col_names_, context);
      }
    } else if (auto x = std::dynamic_pointer_cast<JoinPlan>(plan)) {
      std::unique_ptr<AbstractExecutor> left = convert_plan_executor(x->left_, context);
      std::unique_ptr<AbstractExecutor> right = convert_plan_executor(x->right_, context);
      if (x->tag == T_NestLoop) {
        return std::make_unique<NestedLoopJoinExecutor>(std::move(left), std::move(right),
                                                        std::move(x->conds_));
      } else {
        return std::make_unique<SortMergeJoinExecutor>(std::move(left), std::move(right),
                                                       std::move(x->conds_), x->need_reverse_, x->need_sleep_);
      }
      // std::unique_ptr<AbstractExecutor> join = std::make_unique<NestedLoopJoinExecutor>(
      //     std::move(left), std::move(right), std::move(x->conds_));
      // return join;
    } else if (auto x = std::dynamic_pointer_cast<SortPlan>(plan)) {
      return std::make_unique<SortExecutor>(convert_plan_executor(x->subplan_, context),
                                            x->sel_col_, x->is_desc_);
    } else if (auto x = std::dynamic_pointer_cast<AggregationAndGroupPlan>(plan)) {
      std::unique_ptr<AbstractExecutor> child = convert_plan_executor(x->subplan_, context);
      return std::make_unique<AggregationAndGroupExecutor>(std::move(child), x->group_cols_,
                                                           x->aggregation_cols_, x->having_conds_);
    }
    return nullptr;
  }
};