/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <condition_variable>
#include <list>
#include <memory>
#include <mutex>
#include <unordered_map>

#include "common/config.h"
#include "transaction/transaction.h"
#include "transaction/txn_defs.h"

static const std::string GroupLockModeStr[10] = {"NON_LOCK", "IS", "IX", "S", "X", "SIX"};

// 注意：这里的所有代码都基于“事务均为可串行化的隔离级别”
// 称会在rmdb.cpp中调用main函数的线程为主线程
// 而且，同一事务不可能被多个线程访问，除主线程外只可能被一个线程访问。

// 对wait-die的理解：
// 1.基本：一个资源被事务T1锁定后，只有时间戳小于T1的事务才可以等待，否则abort。
// 2.加入锁的相容：一个资源被事务T1锁定后，只有时间戳小于T1的事务才可以等待，否则abort。
//     如果时间戳小于事务T1，且锁相容，则可以直接给锁；锁不容，则等待。
// 3.加入锁升级：一个资源被事务T1，T1锁定，且TS1 < TS2。
//     如果T1锁升级，那么T1等待T2。如果T2进行锁升级，那么T2 abort。
// 4.加入多个事务共享：一个资源被相容的几个事务T1,T2...Tn锁定后，只有时间戳小于T1的事务才可以等待，否则abort。
//     如果时间戳小于所有事务T1,T2...Tn的时间戳，且锁相容，则可以直接给锁；锁不容，则等待。
// 5.wait-die似乎只有kill itself，而没有let's kill it.
// 6.事务进行锁升级相当于解除了原来的占用，然后重新进行申请，因此该被abort还是会被abort.
// 7.如果多个事务等待，当资源释放时，可根据时间戳大小决定给与资源的顺序。
// 8.一个事物等待，必须时间戳小于队列中所有的事务.
// 9.wait-die似乎是对排斥锁之间进行管理，对于相容锁允许直接获取，不用

// 定义一种表示： 锁类型(时间戳), 比如： S(1) 表示时间戳为1的事务申请了一个S锁。
// [锁类型(时间戳)]表示资源现在被占用的类型。
// 其中，如果时间戳如果为 ? ,表示时间戳是多少并不重要

class LockManager {
  /* 加锁类型，包括共享锁、排他锁、意向共享锁、意向排他锁、SIX（意向排他锁+共享锁）、间隙锁 */
  // 我说为什么我的插件老警告，原来独享锁打错了 exclusive 打成了exlucsive.，算了，懒得管。
  enum class LockMode {
    SHARED,
    EXLUCSIVE,
    INTENTION_SHARED,
    INTENTION_EXCLUSIVE,
    S_IX,
    INDEX_GAP_S,
    Index_GAP_X
  };

  /* 用于标识加锁队列中排他性最强的锁类型，例如加锁队列中有SHARED和EXLUSIVE两个加锁操作，则该队列的锁模式为X
   */
  enum class GroupLockMode { NON_LOCK, IS, IX, S, X, SIX };
  /* 事务的加锁申请 */
  class LockRequest {
   public:
    LockRequest(txn_id_t txn_id, LockMode lock_mode)
        : txn_id_(txn_id), lock_mode_(lock_mode), granted_(false) {}

    txn_id_t txn_id_;     // 申请加锁的事务ID
    LockMode lock_mode_;  // 事务申请加锁的类型
    bool granted_;        // 该事务是否已经被赋予锁
  };

  /* 间隙锁申请 */
  class gapLockRequest {
   public:
    gapLockRequest(txn_id_t txn_id, LockMode lock_mode, Value left, Value right)
        : txn_id_(txn_id),
          lock_mode_(lock_mode),
          low_key_(left),
          upper_key_(right),
          granted_(false) {}

    txn_id_t txn_id_;     // 申请加锁的事务ID
    LockMode lock_mode_;  // 事务申请加锁的类型
    Value low_key_, upper_key_;
    bool granted_;  // 该事务是否已经被赋予锁
  };

  class lockGapList {
   public:
    std::list<std::shared_ptr<gapLockRequest>> request_gap_queue_;
  };

  /* 数据项上的加锁队列 */
  // 队列中只有已经申请到锁的事务和可以等待的事务
  // 不能等待的事务已经被abort。
  // 队列等待队列的顺序请求到达顺序，。
  // 已经申请到资源的事务的队列之间无序。
  class LockRequestQueue {
   public:
    std::list<std::shared_ptr<LockRequest>> request_queue_;  // 加锁队列
    std::list<std::shared_ptr<LockRequest>> locked_queue_;   // 已经申请到的
    std::condition_variable cv_;  // 条件变量，用于唤醒正在等待加锁的申请，在no-wait策略下无需使用
                                  // GroupLockMode group_lock_mode_ =
    // GroupLockMode::NON_LOCK;  // 加锁队列的锁模式, 当前资源的相容的锁的最高级别。
    std::mutex latch_;  // 申请队列的资源的锁，对这些线程来进行管理。
    // 间隙锁使用
    bool is_gap_lock_ = false;
    std::unordered_map<txn_id_t, std::vector<std::string>> txn_gap_cols;  // 每个事务对应的gap_cols
    std::unordered_map<std::string, lockGapList> gap_list_s,
        gap_list_x;  // 每个读、写操作所给加上的间隙锁
  };

 public:
  LockManager() {}

  ~LockManager() {}
  // 间隙锁相关函数
  bool check_gap_lock_timep(LockRequestQueue& lock_queue, txn_id_t t_id, LockMode lock_mode);
  bool check_gap_lock(LockRequestQueue& lock_queue, txn_id_t t_id, LockMode lock_mode,
                      std::string col_name, Value left_val, Value right_val, bool star_wait_die);
  bool lock_gap(Transaction* txn, LockDataId lock_data, LockMode lock_mode,
                std::vector<std::string> col_names, std::vector<Value> insert_values);
  bool lock_gap_on_index(Transaction* txn, int tab_fd, int index_fd,
                         std::vector<std::string> col_names, std::vector<Value> insert_values);
  bool add_index_gap_lock(Transaction* txn, int tab_fd, int index_fd,
                          std::vector<std::string> index_col_names_,
                          std::unordered_map<std::string, Value> left_,
                          std::unordered_map<std::string, Value> right_);
  //
  bool lock_shared_on_record(Transaction* txn, const Rid& rid, int tab_fd);

  bool lock_exclusive_on_record(Transaction* txn, const Rid& rid, int tab_fd);

  bool lock_shared_on_table(Transaction* txn, int tab_fd);

  bool lock_exclusive_on_table(Transaction* txn, int tab_fd);

  bool lock_IS_on_table(Transaction* txn, int tab_fd);

  bool lock_IX_on_table(Transaction* txn, int tab_fd);

  bool unlock(Transaction* txn, LockDataId lock_data_id);

 private:
  // 这个函数不会检查时间戳
  bool grant_lock(LockRequestQueue& lock_queue, txn_id_t t_id, LockMode lock_mode,
                  LockRequest& lock_request);

  // 请求不用等待：与现存锁全兼容
  // 请求不能等待：锁与现存锁不全兼容，并且时间戳不全小于现存事务
  // 请求可以等待：锁与现存锁不全兼容，并且时间戳全小于现存事务，并且与等待队列中的请求要么兼容要么时间戳小于。
  bool check_timep(LockRequestQueue& lock_queue, txn_id_t t_id, LockMode lock_mode);

  // 检查锁是否相容。
  bool check_lock(LockMode left_lock, LockMode right_mode);

  // lock_mode1到lock_mode2是否是锁升级。
  bool is_upgrade(LockMode lock_mode1, LockMode lock_mode2);
  bool lock(Transaction* txn, LockDataId lock_data, LockMode lock_mode);

 private:
  std::mutex latch_;                                             // 用于锁表的并发
  std::unordered_map<LockDataId, LockRequestQueue> lock_table_;  // 全局锁表
};
