/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "transaction_manager.h"

#include <atomic>
#include <condition_variable>
#include <cstddef>
#include <memory>
#include <mutex>
#include <stdexcept>
#include <unordered_set>

#include "common/config.h"
#include "common/log.h"
#include "record/rm_defs.h"
#include "record/rm_file_handle.h"
#include "record/rm_manager.h"
#include "recovery/log_manager.h"
#include "storage/page.h"
#include "system/sm_manager.h"
#include "transaction/transaction.h"
#include "transaction/txn_defs.h"

std::unordered_map<txn_id_t, Transaction*> TransactionManager::txn_map = {};
std::unordered_set<txn_id_t> TransactionManager::ATT = {};
std::atomic<bool> TransactionManager::allow_begin = true;
std::mutex TransactionManager::latch = {};
std::condition_variable TransactionManager::begin_cv = {};
std::condition_variable TransactionManager::wait_cv = {};
std::mutex TransactionManager::check_point_latch = {};

/**
 * @description: 事务的开始方法
 * @return {Transaction*} 开始事务的指针
 * @param {Transaction*} txn 事务指针，空指针代表需要创建新事务，否则开始已有事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
Transaction* TransactionManager::begin(Transaction* txn, LogManager* log_manager) {
  // Todo:
  // 1. 判断传入事务参数是否为空指针
  // 2. 如果为空指针，创建新事务
  // 3. 把开始事务加入到全局事务表中
  // 4. 返回当前事务指针
  if (txn == nullptr) {
    std::unique_lock<std::mutex> lock(latch);

    while (!allow_begin) {
      begin_cv.wait(lock);
    }

    txn = new Transaction(++next_txn_id_, IsolationLevel::SERIALIZABLE);
    txn_map[next_txn_id_] = txn;

#ifdef RECOVERY_ON
    std::unique_lock<std::mutex> llock(check_point_latch);
    ATT.emplace(txn->get_transaction_id());
#endif
    // 对于事务begin的日志推迟处理，主要是为了隐性事务。
    // BeginLogRecord log_record(txn->get_transaction_id());
    // log_record.prev_lsn_ = txn->get_prev_lsn();
    // lsn_t prev_lsn = log_manager->add_log_to_buffer(&log_record);
    // txn->set_prev_lsn(prev_lsn);
  }
  return txn;
}

/**
 * @description: 事务的提交方法
 * @param {Transaction*} txn 需要提交的事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
void TransactionManager::commit(Transaction* txn, LogManager* log_manager) {
  // Todo:
  // 1. 如果存在未提交的写操作，提交所有的写操作
  // 2. 释放所有锁
  // 3. 释放事务相关资源，eg.锁集
  // 4. 把事务日志刷入磁盘中
  // 5. 更新事务状态

#ifdef RECOVERY_ON
  if (txn->get_has_write()) {
    // 隐性事务不再写入commit日志。
    CommitLogRecord log_record(txn->get_transaction_id());
    log_record.prev_lsn_ = txn->get_prev_lsn();
    lsn_t prev_lsn = log_manager->add_log_to_buffer(&log_record);
    txn->set_prev_lsn(prev_lsn);

    if (txn->get_txn_mode() == EXPLICIT_TXN) {
      log_manager->wait_flush();
    }
  }
#endif

  for (auto& lock : *txn->get_lock_set()) {
    lock_manager_->unlock(txn, lock);
  }

  txn->get_lock_set()->clear();

  // // 在第七题的情况下，没有并发，没有崩溃，那我提交的时候完全可以什么事情也不做。
  txn->set_state(TransactionState::COMMITTED);

  std::unique_lock<std::mutex> lock(check_point_latch);
  ATT.erase(txn->get_transaction_id());
  wait_cv.notify_all();
}

/**
 * @description: 事务的终止（回滚）方法
 * @param {Transaction *} txn 需要回滚的事务
 * @param {LogManager} *log_manager 日志管理器指针
 */
void TransactionManager::abort(Transaction* txn, LogManager* log_manager) {
  // Todo:
  // 1. 回滚所有写操作
  // 2. 释放所有锁
  // 3. 清空事务相关资源，eg.锁集
  // 4. 把事务日志刷入磁盘中
  // 5. 更新事务状态

  // 回滚写操作：
  {
    // aaaaa, undo应该逆序执行，我忘了aaa
    // 之前错误的原因：
    // 使用了rm_manager->open_file()，然后对这个新创建的对象进行操作了。
    // 实际上应该使用sm_manager->fhs_里面的对象，也就是说open_file()应该只有sm_manager去调用。
    // 不然的话，两个对象不一样，可能会导致数据过多的时候表头不一样，最后刷新之后导致读取的数据出错。

    auto write_set = txn->get_write_set();

    for (auto iter = write_set->rbegin(); iter != write_set->rend(); ++iter) {
      auto write_record = *iter;

      auto tab = sm_manager_->fhs_.at(write_record->GetTableName()).get();
      auto& tab_meta = sm_manager_->db_.get_table(write_record->GetTableName());

      switch (write_record->GetWriteType()) {
        case WType::DELETE_TUPLE: {
          auto& rid = write_record->GetRid();
          auto& tuple = write_record->GetRecord();

          InsertLogRecord log_record(txn->get_transaction_id(), tuple, rid,
                                     write_record->GetTableName());
          log_record.prev_lsn_ = txn->get_prev_lsn();
          lsn_t prev_lsn = log_manager->add_log_to_buffer(&log_record);
          txn->set_prev_lsn(prev_lsn);

          // 表插入。
          tab->insert_record(rid, tuple.data, nullptr);

          // 索引插入
          for (auto& ix_meta : tab_meta.indexes) {
            auto ix = sm_manager_->ihs_
                          .at(sm_manager_->get_ix_manager()->get_index_name(
                              write_record->GetTableName(), ix_meta.cols))
                          .get();

            RmRecord index_tuple(ix_meta.col_tot_len);
            ix_meta.make_index_tuple(&tuple, &index_tuple);

            ix->insert_entry(index_tuple.data, rid, txn);
          }
        } break;
        case WType::INSERT_TUPLE: {
          auto& rid = write_record->GetRid();
          auto tuple = tab->get_record(rid, nullptr);

          DeleteLogRecord log_record(txn->get_transaction_id(), *tuple, rid,
                                     write_record->GetTableName());
          log_record.prev_lsn_ = txn->get_prev_lsn();
          lsn_t prev_lsn = log_manager->add_log_to_buffer(&log_record);
          txn->set_prev_lsn(prev_lsn);

          for (auto& ix_meta : tab_meta.indexes) {
            auto ix = sm_manager_->ihs_
                          .at(sm_manager_->get_ix_manager()->get_index_name(
                              write_record->GetTableName(), ix_meta.cols))
                          .get();

            RmRecord index_tuple(ix_meta.col_tot_len);
            ix_meta.make_index_tuple(tuple.get(), &index_tuple);

            ix->delete_entry(index_tuple.data, txn);
          }

          tab->delete_record(rid, nullptr);
        } break;
        case WType::UPDATE_TUPLE: {
          auto& rid = write_record->GetRid();
          auto& old_tuple = write_record->GetRecord();
          auto new_tuple = tab->get_record(rid, nullptr);

          UpdateLogRecord log_record(txn->get_transaction_id(), old_tuple, *new_tuple, rid,
                                     write_record->GetTableName());
          log_record.prev_lsn_ = txn->get_prev_lsn();
          lsn_t prev_lsn = log_manager->add_log_to_buffer(&log_record);
          txn->set_prev_lsn(prev_lsn);

          for (auto& ix_meta : tab_meta.indexes) {
            auto ix = sm_manager_->ihs_
                          .at(sm_manager_->get_ix_manager()->get_index_name(
                              write_record->GetTableName(), ix_meta.cols))
                          .get();

            RmRecord index_tuple(ix_meta.col_tot_len);

            ix_meta.make_index_tuple(new_tuple.get(), &index_tuple);

            ix->delete_entry(index_tuple.data, txn);

            ix_meta.make_index_tuple(&old_tuple, &index_tuple);

            ix->insert_entry(index_tuple.data, rid, txn);
          }

          tab->update_record(rid, old_tuple.data, nullptr);
        } break;
        default: {
          ASSERT(false, "error write type.");
        } break;
      }
    }
  }

#ifdef RECOVERY_ON
  if (txn->get_has_write()) {
    AbortLogRecord log_record(txn->get_transaction_id());
    log_record.prev_lsn_ = txn->get_prev_lsn();
    lsn_t prev_lsn = log_manager->add_log_to_buffer(&log_record);
    txn->set_prev_lsn(prev_lsn);

    if (txn->get_txn_mode() == EXPLICIT_TXN) {
      log_manager->wait_flush();
    }
  }
#endif

  for (auto& lock : *txn->get_lock_set()) {
    lock_manager_->unlock(txn, lock);
  }
  txn->get_lock_set()->clear();

  txn->set_state(TransactionState::ABORTED);

  std::unique_lock<std::mutex> lock(check_point_latch);
  ATT.erase(txn->get_transaction_id());
  wait_cv.notify_all();
}